
#undef FL
#ifndef HAVE_AMALGAMATION
# define FL                      extern
#else
# define FL                      static
#endif

/* Memory allocation routines from memory.c offer some debug support */
#if defined HAVE_DEBUG || defined HAVE_DEVEL
# define HAVE_MEMORY_DEBUG
# define SMALLOC_DEBUG_ARGS      , char const *mdbg_file, int mdbg_line
# define SMALLOC_DEBUG_ARGSCALL  , mdbg_file, mdbg_line
# define SALLOC_DEBUG_ARGS       , char const *mdbg_file, int mdbg_line
# define SALLOC_DEBUG_ARGSCALL   , mdbg_file, mdbg_line
#else
# define SMALLOC_DEBUG_ARGS
# define SMALLOC_DEBUG_ARGSCALL
# define SALLOC_DEBUG_ARGS
# define SALLOC_DEBUG_ARGSCALL
#endif

/*
 * Macro-based generics
 */

/* ASCII char classification */
#define __ischarof(C, FLAGS)  \
   (asciichar(C) && (class_char[(uc_i)(C)] & (FLAGS)) != 0)

#define asciichar(c)    ((uc_i)(c) <= 0177)
#define alnumchar(c)    __ischarof(c, C_DIGIT | C_OCTAL | C_UPPER | C_LOWER)
#define alphachar(c)    __ischarof(c, C_UPPER | C_LOWER)
#define blankchar(c)    __ischarof(c, C_BLANK)
#define blankspacechar(c) __ischarof(c, C_BLANK | C_SPACE)
#define cntrlchar(c)    __ischarof(c, C_CNTRL)
#define digitchar(c)    __ischarof(c, C_DIGIT | C_OCTAL)
#define lowerchar(c)    __ischarof(c, C_LOWER)
#define punctchar(c)    __ischarof(c, C_PUNCT)
#define spacechar(c)    __ischarof(c, C_BLANK | C_SPACE | C_WHITE)
#define upperchar(c)    __ischarof(c, C_UPPER)
#define whitechar(c)    __ischarof(c, C_BLANK | C_WHITE)
#define octalchar(c)    __ischarof(c, C_OCTAL)

#define upperconv(c)    (lowerchar(c) ? (char)((uc_i)(c) - 'a' + 'A') : (c))
#define lowerconv(c)    (upperchar(c) ? (char)((uc_i)(c) - 'A' + 'a') : (c))
/* RFC 822, 3.2. */
#define fieldnamechar(c) \
   (asciichar(c) && (c) > 040 && (c) != 0177 && (c) != ':')

/* Could the string contain a regular expression? */
#if 0
# define is_maybe_regex(S) anyof("^.[]*+?()|$", S)
#else
# define is_maybe_regex(S) anyof("^[]*+?|$", S)
#endif

/* Single-threaded, use unlocked I/O */
#ifdef HAVE_PUTC_UNLOCKED
# undef getc
# define getc(c)        getc_unlocked(c)
# undef putc
# define putc(c, f)     putc_unlocked(c, f)
# undef putchar
# define putchar(c)     putc_unlocked((c), stdout)
#endif

/* Truncate a file to the last character written.  This is useful just before
 * closing an old file that was opened for read/write */
#define ftrunc(stream) \
do {\
   off_t off;\
   fflush(stream);\
   off = ftell(stream);\
   if (off >= 0)\
      ftruncate(fileno(stream), off);\
} while (0)

/* fflush() and rewind() */
#define fflush_rewind(stream) \
do {\
   fflush(stream);\
   rewind(stream);\
} while (0)

/* There are problems with dup()ing of file-descriptors for child processes.
 * As long as those are not fixed in equal spirit to (outof(): FIX and
 * recode.., 2012-10-04), and to avoid reviving of bugs like (If *record* is
 * set, avoid writing dead content twice.., 2012-09-14), we have to somehow
 * accomplish that the FILE* fp makes itself comfortable with the *real* offset
 * of the underlaying file descriptor.  Unfortunately Standard I/O and POSIX
 * don't describe a way for that -- fflush();rewind(); won't do it.  This
 * fseek(END),rewind() pair works around the problem on *BSD and Linux.
 * Update as of 2014-03-03: with Issue 7 POSIX has overloaded fflush(3): if
 * used on a readable stream, then
 *
 *    if the file is not already at EOF, and the file is one capable of
 *    seeking, the file offset of the underlying open file description shall
 *    be set to the file position of the stream.
 *
 * We need our own, simplified and reliable I/O */
#if defined _POSIX_VERSION && _POSIX_VERSION + 0 >= 200809L
# define really_rewind(stream) \
do {\
   rewind(stream);\
   fflush(stream);\
} while (0)
#else
# define really_rewind(stream) \
do {\
   fseek(stream, 0, SEEK_END);\
   rewind(stream);\
} while (0)
#endif

/*
 * accmacvar.c
 */

/* Don't use _var_* unless you *really* have to! */

/* Constant option key look/(un)set/clear */
FL char *      _var_oklook(enum okeys okey);
#define ok_blook(C)              (_var_oklook(CONCAT(ok_b_, C)) != NULL)
#define ok_vlook(C)              _var_oklook(CONCAT(ok_v_, C))

FL bool_t      _var_okset(enum okeys okey, uintptr_t val);
#define ok_bset(C,B)             _var_okset(CONCAT(ok_b_, C), (uintptr_t)(B))
#define ok_vset(C,V)             _var_okset(CONCAT(ok_v_, C), (uintptr_t)(V))

FL bool_t      _var_okclear(enum okeys okey);
#define ok_bclear(C)             _var_okclear(CONCAT(ok_b_, C))
#define ok_vclear(C)             _var_okclear(CONCAT(ok_v_, C))

/* Variable option key look/(un)set/clear */
FL char *      _var_voklook(char const *vokey);
#define vok_blook(S)              (_var_voklook(S) != NULL)
#define vok_vlook(S)              _var_voklook(S)

FL bool_t      _var_vokset(char const *vokey, uintptr_t val);
#define vok_bset(S,B)            _var_vokset(S, (uintptr_t)(B))
#define vok_vset(S,V)            _var_vokset(S, (uintptr_t)(V))

FL bool_t      _var_vokclear(char const *vokey);
#define vok_bclear(S)            _var_vokclear(S)
#define vok_vclear(S)            _var_vokclear(S)

/* Environment lookup, if envonly is TRU1 then variable must come from the
 * process environment (and if via `setenv') */
FL char *      _env_look(char const *envkey, bool_t envonly);
#define env_blook(S,EXCL)        (_env_look(S, EXCL) != NULL)
#define env_vlook(S,EXCL)        _env_look(S, EXCL)

/* Special case to handle the typical [xy-USER@HOST,] xy-HOST and plain xy
 * variable chains; oxm is a bitmix which tells which combinations to test */
#ifdef HAVE_SOCKETS
FL char *      _var_xoklook(enum okeys okey, struct url const *urlp,
                  enum okey_xlook_mode oxm);
# define xok_BLOOK(C,URL,M)      (_var_xoklook(C, URL, M) != NULL)
# define xok_VLOOK(C,URL,M)      _var_xoklook(C, URL, M)
# define xok_blook(C,URL,M)      xok_BLOOK(CONCAT(ok_b_, C), URL, M)
# define xok_vlook(C,URL,M)      xok_VLOOK(CONCAT(ok_v_, C), URL, M)
#endif

/* `varshow' */
FL int         c_varshow(void *v);

/* User variable access: `set', `setenv', `unset' and `unsetenv' */
FL int         c_set(void *v);
FL int         c_setenv(void *v);
FL int         c_unset(void *v);
FL int         c_unsetenv(void *v);

/* Ditto: `varedit' */
FL int         c_varedit(void *v);

/* Macros: `define', `undefine', `call' / `~' */
FL int         c_define(void *v);
FL int         c_undefine(void *v);
FL int         c_call(void *v);

/* Check wether a *folder-hook* exists for the currently active mailbox */
FL bool_t      check_folder_hook(bool_t nmail);

/* Accounts: `account', `unaccount' */
FL int         c_account(void *v);
FL int         c_unaccount(void *v);

/* `localopts' */
FL int         c_localopts(void *v);

FL void        temporary_localopts_free(void); /* XXX intermediate hack */
FL void        temporary_localopts_folder_hook_unroll(void); /* XXX im. hack */

/*
 * attachments.c
 */

/* Try to add an attachment for file, file_expand()ed.
 * Return the new head of list aphead, or NULL.
 * The newly created attachment will be stored in *newap, if given */
FL struct attachment * add_attachment(struct attachment *aphead, char *file,
                        struct attachment **newap);

/* Append comma-separated list of file names to the end of attachment list */
FL void        append_attachments(struct attachment **aphead, char *names);

/* Interactively edit the attachment list */
FL void        edit_attachments(struct attachment **aphead);

/*
 * auxlily.c
 */

/* Compute screen size */
FL int         screensize(void);

/* Get our $PAGER; if env_addon is not NULL it is checked wether we know about
 * some environment variable that supports colour+ and set *env_addon to that,
 * e.g., "LESS=FRSXi" */
FL char const *n_pager_get(char const **env_addon);

/* Use a pager or STDOUT to print *fp*; if *lines* is 0, they'll be counted */
FL void        page_or_print(FILE *fp, size_t lines);

/* Parse name and guess at the required protocol */
FL enum protocol  which_protocol(char const *name);

/* Hash the passed string -- uses Chris Torek's hash algorithm */
FL ui32_t      torek_hash(char const *name);
#define hash(S)                  (torek_hash(S) % HSHSIZE) /* xxx COMPAT (?) */

/* Create hash */
FL ui32_t      pjw(char const *cp); /* TODO obsolete -> torek_hash() */

/* Find a prime greater than n */
FL ui32_t      nextprime(ui32_t n);

/* Get *prompt*, or '& ' if *bsdcompat*, of '? ' otherwise */
FL char *      getprompt(void);

/* Detect and query the hostname to use */
FL char *      nodename(int mayoverride);

/* Get a (pseudo) random string of *length* bytes; returns salloc()ed buffer */
FL char *      getrandstring(size_t length);

/* Check wether the argument string is a true (1) or false (0) boolean, or an
 * invalid string, in which case -1 is returned; if emptyrv is not -1 then it,
 * treated as a boolean, is used as the return value shall inbuf be empty.
 * inlen may be UIZ_MAX to force strlen() detection */
FL si8_t       boolify(char const *inbuf, uiz_t inlen, si8_t emptyrv);

/* Dig a "quadoption" in inbuf (possibly going through getapproval() in
 * interactive mode).  Returns a boolean or -1 if inbuf content is invalid;
 * if emptyrv is not -1 then it,  treated as a boolean, is used as the return
 * value shall inbuf be empty.  If prompt is set it is printed first if intera.
 * inlen may be UIZ_MAX to force strlen() detection */
FL si8_t       quadify(char const *inbuf, uiz_t inlen, char const *prompt,
                  si8_t emptyrv);

/* Get seconds since epoch */
FL time_t      n_time_epoch(void);

/* Update *tc* to now; only .tc_time updated unless *full_update* is true */
FL void        time_current_update(struct time_current *tc,
                  bool_t full_update);

/* Our error print series..  Note: these reverse scan format in order to know
 * wether a newline was included or not -- this affects the output! */
FL void        n_err(char const *format, ...);
FL void        n_verr(char const *format, va_list ap);

/* ..(for use in a signal handler; to be obsoleted..).. */
FL void        n_err_sighdl(char const *format, ...);

/* Our perror(3); if errval is 0 errno(3) is used; newline appended */
FL void        n_perr(char const *msg, int errval);

/* Announce a fatal error (and die); newline appended */
FL void        n_alert(char const *format, ...);
FL void        n_panic(char const *format, ...);

/* `errors' */
#ifdef HAVE_ERRORS
FL int         c_errors(void *vp);
#else
# define c_errors                c_cmdnotsupp
#endif

/*
 * cmd_cnd.c
 */

/* if.elif.else.endif conditional execution.
 * condstack_isskip() returns wether the current condition state doesn't allow
 * execution of commands.
 * condstack_release() and condstack_take() rotate the current condition stack;
 * condstack_take() returns a false boolean if the current condition stack has
 * unclosed conditionals */
FL int         c_if(void *v);
FL int         c_elif(void *v);
FL int         c_else(void *v);
FL int         c_endif(void *v);
FL bool_t      condstack_isskip(void);
FL void *      condstack_release(void);
FL bool_t      condstack_take(void *self);

/*
 * collect.c
 */

FL FILE *      collect(struct header *hp, int printheaders, struct message *mp,
                  char *quotefile, int doprefix, si8_t *checkaddr_err);

FL void        savedeadletter(FILE *fp, int fflush_rewind_first);

/*
 * edit.c
 */

/* Edit a message list */
FL int         c_editor(void *v);

/* Invoke the visual editor on a message list */
FL int         c_visual(void *v);

/* Run an editor on either size bytes of the file fp (or until EOF if size is
 * negative) or on the message mp, and return a new file or NULL on error of if
 * the user didn't perform any edits.
 * For now we assert that mp==NULL if hp!=NULL, treating this as a special call
 * from within compose mode, and giving TRUM1 to puthead().
 * Signals must be handled by the caller.  viored is 'e' for ed, 'v' for vi */
FL FILE *      run_editor(FILE *fp, off_t size, int viored, int readonly,
                  struct header *hp, struct message *mp,
                  enum sendaction action, sighandler_type oldint);

/*
 * fio.c
 */

/* fgets() replacement to handle lines of arbitrary size and with embedded \0
 * characters.
 * line - line buffer.  *line may be NULL.
 * linesize - allocated size of line buffer.
 * count - maximum characters to read.  May be NULL.
 * llen - length_of_line(*line).
 * fp - input FILE.
 * appendnl - always terminate line with \n, append if necessary.
 */
FL char *      fgetline(char **line, size_t *linesize, size_t *count,
                  size_t *llen, FILE *fp, int appendnl SMALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define fgetline(A,B,C,D,E,F)   \
   fgetline(A, B, C, D, E, F, __FILE__, __LINE__)
#endif

/* Read up a line from the specified input into the linebuffer.
 * Return the number of characters read.  Do not include the newline at EOL.
 * n is the number of characters already read */
FL int         readline_restart(FILE *ibuf, char **linebuf, size_t *linesize,
                  size_t n SMALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define readline_restart(A,B,C,D) \
   readline_restart(A, B, C, D, __FILE__, __LINE__)
#endif

/* Set up the input pointers while copying the mail file into /tmp */
FL void        setptr(FILE *ibuf, off_t offset);

/* Drop the passed line onto the passed output buffer.  If a write error occurs
 * return -1, else the count of characters written, including the newline */
FL int         putline(FILE *obuf, char *linebuf, size_t count);

/* Determine the size of the file possessed by the passed buffer */
FL off_t       fsize(FILE *iob);

/* accmacvar.c hook: *folder* variable has been updated; if folder shouldn't
 * be replaced by something else leave store alone, otherwise smalloc() the
 * desired value (ownership will be taken) */
FL bool_t      var_folder_updated(char const *folder, char **store);

/* Return the name of the dead.letter file */
FL char const * getdeadletter(void);

/* Will retry FILE_LOCK_RETRIES times if pollmsecs > 0 */
FL bool_t      n_file_lock(int fd, enum n_file_lock_type flt,
                  off_t off, off_t len, size_t pollmsecs);

/*
 * lex_input.c
 */

/* Print the docstring of `comm', which may be an abbreviation.
 * Return FAL0 if there is no such command */
#ifdef HAVE_DOCSTRINGS
FL bool_t      print_comm_docstr(char const *comm);
#endif

/* Interpret user commands.  If standard input is not a tty, print no prompt;
 * return wether the last processed command returned error */
FL bool_t      commands(void);
FL bool_t      commands_recursive(void);

/* Actual cmd input */

/* Read a complete line of input, with editing if interactive and possible.
 * If prompt is NULL we'll call getprompt() first, if necessary.
 * nl_escape defines wether user can escape newlines via backslash (POSIX).
 * If string is set it is used as the initial line content if in interactive
 * mode, otherwise this argument is ignored for reproducibility.
 * Return number of octets or a value <0 on error.
 * Note: may use the currently `source'd file stream instead of stdin! */
FL int         n_lex_input(char const *prompt, bool_t nl_escape,
                  char **linebuf, size_t *linesize, char const *string
                  SMALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define n_lex_input(A,B,C,D,E) n_lex_input(A,B,C,D,E,__FILE__,__LINE__)
#endif

/* Read a line of input, with editing if interactive and possible, return it
 * savestr()d or NULL in case of errors or if an empty line would be returned.
 * This may only be called from toplevel (not during PS_ROBOT).
 * If prompt is NULL we'll call getprompt() if necessary.
 * If string is set it is used as the initial line content if in interactive
 * mode, otherwise this argument is ignored for reproducibility.
 * If OPT_INTERACTIVE a non-empty return is saved in the history, isgabby */
FL char *      n_lex_input_cp_addhist(char const *prompt, char const *string,
                  bool_t isgabby);

/* Deal with loading of resource files and dealing with a stack of files for
 * the source command */

/* Load a file of user definitions -- this is *only* for main()! */
FL void        load(char const *name);

/* "Load" all the -X command line definitions in order -- *only* for main() */
FL void        load_Xargs(char const **lines);

/* Pushdown current input file and switch to a new one.  Set the global flag
 * PS_SOURCING so that others will realize that they are no longer reading from
 * a tty (in all probability).
 * The latter won't return failure (TODO should be replaced by "-f FILE") */
FL int         c_source(void *v);
FL int         c_source_if(void *v);

/* Evaluate a complete macro / a single command.  For the former lines will
 * always be free()d, for the latter cmd will always be duplicated internally */
FL bool_t      n_source_macro(char const *name, char **lines);
FL bool_t      n_source_command(char const *cmd);

/*
 * main.c
 */

/* Quit quickly.  If PS_ROBOT, just pop the input level by returning error */
FL int         c_rexit(void *v);

/*
 * memory.c
 */

/* Try to use alloca() for some function-local buffers and data, fall back to
 * smalloc()/free() if not available */

#ifdef HAVE_ALLOCA
# define ac_alloc(n)    HAVE_ALLOCA(n)
# define ac_free(n)     do {UNUSED(n);} while (0)
#else
# define ac_alloc(n)    smalloc(n)
# define ac_free(n)     free(n)
#endif

/* Generic heap memory */

FL void *      smalloc(size_t s SMALLOC_DEBUG_ARGS);
FL void *      srealloc(void *v, size_t s SMALLOC_DEBUG_ARGS);
FL void *      scalloc(size_t nmemb, size_t size SMALLOC_DEBUG_ARGS);

#ifdef HAVE_MEMORY_DEBUG
FL void        sfree(void *v SMALLOC_DEBUG_ARGS);

/* Called by sreset(), then */
FL void        n_memreset(void);

FL int         c_memtrace(void *v);

/* For immediate debugging purposes, it is possible to check on request */
FL bool_t      n__memcheck(char const *file, int line);

# define smalloc(SZ)             smalloc(SZ, __FILE__, __LINE__)
# define srealloc(P,SZ)          srealloc(P, SZ, __FILE__, __LINE__)
# define scalloc(N,SZ)           scalloc(N, SZ, __FILE__, __LINE__)
# define free(P)                 sfree(P, __FILE__, __LINE__)
# define c_memtrace              c_memtrace
# define n_memcheck()            n__memcheck(__FILE__, __LINE__)
#else
# define n_memreset()            do{}while(0)
#endif

/* String storage, auto-reclaimed after execution level is left */

/* Allocate size more bytes of space and return the address of the first byte
 * to the caller.  An even number of bytes are always allocated so that the
 * space will always be on a word boundary */
FL void *      salloc(size_t size SALLOC_DEBUG_ARGS);
FL void *      csalloc(size_t nmemb, size_t size SALLOC_DEBUG_ARGS);
#if defined HAVE_DEBUG || defined HAVE_DEVEL
# define salloc(SZ)              salloc(SZ, __FILE__, __LINE__)
# define csalloc(NM,SZ)          csalloc(NM, SZ, __FILE__, __LINE__)
#endif

/* Auto-reclaim string storage; if only_if_relaxed is true then only perform
 * the reset when a srelax_hold() is currently active */
FL void        sreset(bool_t only_if_relaxed);

/* The "problem" with sreset() is that it releases all string storage except
 * what was present once spreserve() had been called; it therefore cannot be
 * called from all that code which yet exists and walks about all the messages
 * in order, e.g. quit(), searches, etc., because, unfortunately, these code
 * paths are reached with new intermediate string dope already in use.
 * Thus such code should take a srelax_hold(), successively call srelax() after
 * a single message has been handled, and finally srelax_rele() (unless it is
 * clear that sreset() occurs anyway) */
FL void        srelax_hold(void);
FL void        srelax_rele(void);
FL void        srelax(void);

/* Make current string storage permanent: new allocs will be auto-reclaimed by
 * sreset().  This is called once only, from within main() */
FL void        spreserve(void);

/* 'sstats' command */
#if defined HAVE_DEBUG || defined HAVE_DEVEL
FL int         c_sstats(void *v);
# define c_sstats                c_sstats
#endif


/*
 * popen.c
 * Subprocesses, popen, but also file handling with registering
 */

/* For program startup in main.c: initialize process manager */
FL void        command_manager_start(void);

/* Notes: OF_CLOEXEC is implied in oflags, xflags may be NULL */
FL FILE *      safe_fopen(char const *file, char const *oflags, int *xflags);

/* Notes: OF_CLOEXEC|OF_REGISTER are implied in oflags!
 * Exception is Fdopen() if nocloexec is TRU1, but otherwise even for it the fd
 * creator has to take appropriate steps in order to ensure this is true! */
FL FILE *      Fopen(char const *file, char const *oflags);
FL FILE *      Fdopen(int fd, char const *oflags, bool_t nocloexec);

FL int         Fclose(FILE *fp);

/* Open file according to oflags (see popen.c).  Handles compressed files */
FL FILE *      Zopen(char const *file, char const *oflags);

/* Create a temporary file in tempdir, use namehint for its name (prefix
 * unless OF_SUFFIX is set, in which case namehint is an extension that MUST be
 * part of the resulting filename, otherwise Ftmp() will fail), store the
 * unique name in fn (unless OF_UNLINK is set in oflags), and return a stdio
 * FILE pointer with access oflags.  OF_CLOEXEC is implied in oflags.
 * One of OF_WRONLY and OF_RDWR must be set.  Mode of 0600 is implied */
FL FILE *      Ftmp(char **fn, char const *namehint, enum oflags oflags);

/* If OF_HOLDSIGS was set when calling Ftmp(), then hold_all_sigs() had been
 * called: call this to unlink(2) and free *fn and to rele_all_sigs() */
FL void        Ftmp_release(char **fn);

/* Free the resources associated with the given filename.  To be called after
 * unlink() */
FL void        Ftmp_free(char **fn);

/* Create a pipe and ensure CLOEXEC bit is set in both descriptors */
FL bool_t      pipe_cloexec(int fd[2]);

/*
 * env_addon may be NULL, otherwise it is expected to be a NULL terminated
 * array of "K=V" strings to be placed into the childs environment.
 * If cmd==(char*)-1 then shell is indeed expected to be a PTF :P that will be
 * called from within the child process */
FL FILE *      Popen(char const *cmd, char const *mode, char const *shell,
                  char const **env_addon, int newfd1);
FL bool_t      Pclose(FILE *fp, bool_t dowait);

/* In OPT_INTERACTIVE, we want to go over $PAGER.
 * These are specialized version of Popen()/Pclose() which also encapsulate
 * error message printing, terminal handling etc. additionally */
FL FILE *      n_pager_open(void);
FL bool_t      n_pager_close(FILE *fp);

/*  */
FL void        close_all_files(void);

/* Fork a child process, enable use of the *child() series below */
FL int         fork_child(void);

/* Run a command without a shell, with optional arguments and splicing of stdin
 * and stdout.  FDs may also be COMMAND_FD_NULL and COMMAND_FD_PASS, meaning to
 * redirect from/to /dev/null or pass through our own set of FDs; in the
 * latter case terminal capabilities are saved/restored if possible.
 * The command name can be a sequence of words.
 * Signals must be handled by the caller.  "Mask" contains the signals to
 * ignore in the new process.  SIGINT is enabled unless it's in the mask.
 * env_addon may be NULL, otherwise it is expected to be a NULL terminated
 * array of "K=V" strings to be placed into the childs environment */
FL int         run_command(char const *cmd, sigset_t *mask, int infd,
                  int outfd, char const *a0, char const *a1, char const *a2,
                  char const **env_addon);

/* Like run_command, but don't wait for the command to finish.
 * Also it is usually an error to use COMMAND_FD_PASS for this one */
FL int         start_command(char const *cmd, sigset_t *mask, int infd,
                  int outfd, char const *a0, char const *a1, char const *a2,
                  char const **env_addon);

/* In-child process */
FL void        prepare_child(sigset_t *nset, int infd, int outfd);

/* Mark a child as don't care - pid */
FL void        free_child(int pid);

/* Wait for pid, return wether we've had a normal EXIT_SUCCESS exit.
 * If wait_status is set, set it to the reported waitpid(2) wait status */
FL bool_t      wait_child(int pid, int *wait_status);

/*
 * signal.c
 */

FL void        n_raise(int signo);

/* Provide BSD-like signal() on all (POSIX) systems */
FL sighandler_type safe_signal(int signum, sighandler_type handler);

/* Hold *all* signals but SIGCHLD, and release that total block again */
FL void        hold_all_sigs(void);
FL void        rele_all_sigs(void);

/* Hold HUP/QUIT/INT */
FL void        hold_sigs(void);
FL void        rele_sigs(void);

/* Call _ENTER_SWITCH() with the according flags, it'll take care for the rest
 * and also set the jump buffer - it returns 0 if anything went fine and
 * a signal number if a jump occurred, in which case all handlers requested in
 * flags are temporarily SIG_IGN.
 * _cleanup_ping() informs the condome that no jumps etc. shall be performed
 * until _leave() is called in the following -- to be (optionally) called right
 * before the local jump label is reached which is jumped to after a long jump
 * occurred, straight code flow provided, e.g., to avoid destructors to be
 * called twice.  _leave() must always be called last, reraise_flags will be
 * used to decide how signal handling has to continue
 */
#define n_SIGMAN_ENTER_SWITCH(S,F) do{\
   int __x__;\
   hold_sigs();\
   if(sigsetjmp((S)->sm_jump, 1))\
      __x__ = -1;\
   else\
      __x__ = F;\
   n__sigman_enter(S, __x__);\
}while(0); switch((S)->sm_signo)
FL int         n__sigman_enter(struct n_sigman *self, int flags);
FL void        n_sigman_cleanup_ping(struct n_sigman *self);
FL void        n_sigman_leave(struct n_sigman *self, enum n_sigman_flags flags);

/* Pending signal or 0? */
FL int         n_sigman_peek(void);
FL void        n_sigman_consume(void);

/* Not-Yet-Dead debug information (handler installation in main.c) */
#if defined HAVE_DEBUG || defined HAVE_DEVEL
FL void        _nyd_chirp(ui8_t act, char const *file, ui32_t line,
                  char const *fun);
FL void        _nyd_oncrash(int signo);

# define HAVE_NYD
# define NYD_ENTER               _nyd_chirp(1, __FILE__, __LINE__, __FUN__)
# define NYD_LEAVE               _nyd_chirp(2, __FILE__, __LINE__, __FUN__)
# define NYD                     _nyd_chirp(0, __FILE__, __LINE__, __FUN__)
# define NYD_X                   _nyd_chirp(0, __FILE__, __LINE__, __FUN__)
# ifdef HAVE_NYD2
#  define NYD2_ENTER             _nyd_chirp(1, __FILE__, __LINE__, __FUN__)
#  define NYD2_LEAVE             _nyd_chirp(2, __FILE__, __LINE__, __FUN__)
#  define NYD2                   _nyd_chirp(0, __FILE__, __LINE__, __FUN__)
# endif
#else
# undef HAVE_NYD
#endif
#ifndef NYD
# define NYD_ENTER               do {} while (0)
# define NYD_LEAVE               do {} while (0)
# define NYD                     do {} while (0)
# define NYD_X                   do {} while (0) /* XXX LEGACY */
#endif
#ifndef NYD2
# define NYD2_ENTER              do {} while (0)
# define NYD2_LEAVE              do {} while (0)
# define NYD2                    do {} while (0)
#endif
/*
 * strings.c
 */

/* Return a pointer to a dynamic copy of the argument */
FL char *      savestr(char const *str SALLOC_DEBUG_ARGS);
FL char *      savestrbuf(char const *sbuf, size_t sbuf_len SALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define savestr(CP)             savestr(CP, __FILE__, __LINE__)
# define savestrbuf(CBP,CBL)     savestrbuf(CBP, CBL, __FILE__, __LINE__)
#endif

/* Concatenate cp2 onto cp1 (if not NULL), separated by sep (if not '\0') */
FL char *      savecatsep(char const *cp1, char sep, char const *cp2
                  SALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define savecatsep(S1,SEP,S2)   savecatsep(S1, SEP, S2, __FILE__, __LINE__)
#endif

/* Make copy of argument incorporating old one, if set, separated by space */
#define save2str(S,O)            savecatsep(O, ' ', S)

/* strcat */
#define savecat(S1,S2)           savecatsep(S1, '\0', S2)

/* Create duplicate, lowercasing all characters along the way */
FL char *      i_strdup(char const *src SALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define i_strdup(CP)            i_strdup(CP, __FILE__, __LINE__)
#endif

/*  */
FL struct str * str_concat_csvl(struct str *self, ...);

/*  */
FL struct str * str_concat_cpa(struct str *self, char const * const *cpa,
                  char const *sep_o_null SALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define str_concat_cpa(S,A,N)   str_concat_cpa(S, A, N, __FILE__, __LINE__)
#endif

/* Plain char* support, not auto-reclaimed (unless noted) */

/* Are any of the characters in the two strings the same? */
FL int         anyof(char const *s1, char const *s2);

/* Treat *iolist as a sep separated list of strings; find and return the
 * next entry, trimming surrounding whitespace, and point *iolist to the next
 * entry or to NULL if no more entries are contained.  If ignore_empty is not
 * set empty entries are started over.  Return NULL or an entry */
FL char *      n_strsep(char **iolist, char sep, bool_t ignore_empty);

/* Copy a string, lowercasing it as we go; *size* is buffer size of *dest*;
 * *dest* will always be terminated unless *size* is 0 */
FL void        i_strcpy(char *dest, char const *src, size_t size);

/* Is *as1* a valid prefix of *as2*? */
FL int         is_prefix(char const *as1, char const *as2);

/* Backslash quote (" and \) v'alue, and return salloc()ed result */
FL char *      string_quote(char const *v);

/* Get (and isolate) the last, possibly quoted part of linebuf, set *needs_list
 * to indicate wether getmsglist() et al need to be called to collect
 * additional args that remain in linebuf.  If strip is true possibly
 * surrounding quote characters are removed.  Return NULL on "error" */
FL char *      laststring(char *linebuf, bool_t *needs_list, bool_t strip);

/* Convert a string to lowercase, in-place and with multibyte-aware */
FL void        makelow(char *cp);

/* Is *sub* a substring of *str*, case-insensitive and multibyte-aware? */
FL bool_t      substr(char const *str, char const *sub);

/*  */
FL char *      sstpcpy(char *dst, char const *src);
FL char *      sstrdup(char const *cp SMALLOC_DEBUG_ARGS);
FL char *      sbufdup(char const *cp, size_t len SMALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define sstrdup(CP)             sstrdup(CP, __FILE__, __LINE__)
# define sbufdup(CP,L)           sbufdup(CP, L, __FILE__, __LINE__)
#endif

/* Copy at most dstsize bytes of src to dst and return string length.
 * Returns -E2BIG if dst is not large enough; dst will always be terminated
 * unless dstsize was 0 on entry */
FL ssize_t     n_strscpy(char *dst, char const *src, size_t dstsize);

/* Case-independent ASCII comparisons */
FL int         asccasecmp(char const *s1, char const *s2);
FL int         ascncasecmp(char const *s1, char const *s2, size_t sz);

/* Case-independent ASCII string find s2 in s1, return it or NULL */
FL char const *asccasestr(char const *s1, char const *s2);

/* Case-independent ASCII check wether as2 is the initial substring of as1 */
FL bool_t      is_asccaseprefix(char const *as1, char const *as2);

/* struct str related support funs TODO _cp->_cs! */

/* *self->s* is srealloc()ed */
#define n_str_dup(S, T)          n_str_assign_buf((S), (T)->s, (T)->l)

/* *self->s* is srealloc()ed; if buflen==UIZ_MAX strlen() is called unless buf
 * is NULL; buf may be NULL if buflen is 0 */
FL struct str * n_str_assign_buf(struct str *self,
                  char const *buf, uiz_t buflen SMALLOC_DEBUG_ARGS);
#define n_str_assign(S, T)       n_str_assign_buf(S, (T)->s, (T)->l)
#define n_str_assign_cp(S, CP)   n_str_assign_buf(S, CP, UIZ_MAX)

/* *self->s* is srealloc()ed, *self->l* incremented; if buflen==UIZ_MAX
 * strlen() is called unless buf is NULL; buf may be NULL if buflen is 0 */
FL struct str * n_str_add_buf(struct str *self, char const *buf, uiz_t buflen
                  SMALLOC_DEBUG_ARGS);
#define n_str_add(S, T)          n_str_add_buf(S, (T)->s, (T)->l)
#define n_str_add_cp(S, CP)      n_str_add_buf(S, CP, UIZ_MAX)

#ifdef HAVE_DEBUG
# define n_str_assign_buf(S,B,BL) n_str_assign_buf(S, B, BL, __FILE__, __LINE__)
# define n_str_add_buf(S,B,BL)   n_str_add_buf(S, B, BL, __FILE__, __LINE__)
#endif

/*
 * termcap.c
 * This is a little bit hairy since it provides some stuff even if HAVE_TERMCAP
 * is false due to encapsulation desire
 */

#ifdef n_HAVE_TCAP
/* termcap(3) / xy lifetime handling -- only called if we're OPT_INTERACTIVE
 * but not doing something in OPT_QUICKRUN_MASK */
FL void        n_termcap_init(void);
FL void        n_termcap_destroy(void);

/* Command multiplexer, returns FAL0 on I/O error, TRU1 on success and TRUM1
 * for commands which are not available and have no builtin fallback.
 * For query options the return represents a true value and -1 error.
 * Will return FAL0 directly unless we've been initialized */
FL ssize_t    n_termcap_cmd(enum n_termcap_cmd cmd, ssize_t a1, ssize_t a2);
# define n_termcap_cmdx(CMD)     n_termcap_cmd(CMD, 0, 0)
#endif /* n_HAVE_TCAP */

/* Ease usage a bit */
#ifdef HAVE_TERMCAP
# define n_TERMCAP_SUSPEND()     n_termcap_cmdx(n_TERMCAP_CMD_te)
# define n_TERMCAP_RESUME()      n_termcap_cmdx(n_TERMCAP_CMD_ti)

# define n_TERMCAP_CLEAR_EOS(IFCA) \
      n_termcap_cmdx(n_TERMCAP_CMD_cd | ((IFCA) ? n_TERMCAP_CMD_FLAG_CA : 0))
# define n_TERMCAP_CLEAR_SCREEN(IFCA) \
      n_termcap_cmdx(n_TERMCAP_CMD_cl | ((IFCA) ? n_TERMCAP_CMD_FLAG_CA : 0))
# define n_TERMCAP_CURSOR_HOME(IFCA) \
      n_termcap_cmdx(n_TERMCAP_CMD_ho | ((IFCA) ? n_TERMCAP_CMD_FLAG_CA : 0))
#else
# define n_TERMCAP_SUSPEND()
# define n_TERMCAP_RESUME()

# define n_TERMCAP_CLEAR_EOS(IFCA)
# define n_TERMCAP_CLEAR_SCREEN(IFCA)
# define n_TERMCAP_CURSOR_HOME(IFCA)
#endif

/*
 * tty.c
 */

/* Return wether user says yes, on STDIN if interactive.
 * Uses noninteract_default, the return value for non-interactive use cases,
 * also to choose a default prompt if that is NULL as well as a hint for
 * boolify().  Handles+reraises SIGINT */
FL bool_t      getapproval(char const *prompt, bool_t noninteract_default);

#ifdef HAVE_SOCKETS
/* Get a password the expected way, return termios_state.ts_linebuf on
 * success or NULL on error */
FL char *      getuser(char const *query);

/* Get a password the expected way, return termios_state.ts_linebuf on
 * success or NULL on error.  SIGINT is temporarily blocked, *not* reraised.
 * termios_state_reset() (def.h) must be called anyway */
FL char *      getpassword(char const *query);
#endif

/* Overall interactive terminal life cycle for command line editor library */
#if defined HAVE_READLINE
# define TTY_WANTS_SIGWINCH
#endif
FL void        n_tty_init(void);
FL void        n_tty_destroy(void);

/* Rather for main.c / SIGWINCH interaction only */
FL void        n_tty_signal(int sig);

/* Read a line after printing prompt (if set and non-empty).
 * If n>0 assumes that *linebuf has n bytes of default content */
FL int         n_tty_readline(char const *prompt, char **linebuf,
                  size_t *linesize, size_t n SMALLOC_DEBUG_ARGS);
#ifdef HAVE_DEBUG
# define n_tty_readline(A,B,C,D) n_tty_readline(A, B, C, D, __FILE__, __LINE__)
#endif

/* Add a line (most likely as returned by n_tty_readline()) to the history.
 * Wether an entry added for real depends on the isgabby / *history-gabby*
 * relation, and / or wether s is non-empty and doesn't begin with U+0020 */
FL void        n_tty_addhist(char const *s, bool_t isgabby);

#ifdef HAVE_HISTORY
FL int         c_history(void *v);
#else
# define c_history               c_cmdnotsupp
#endif

/*
 * ui_str.c
 */

/* Detect visual width of (blen bytes of) buf, return (size_t)-1 on error.
 * Give blen UIZ_MAX to strlen().   buf may be NULL if (final) blen is 0 */
FL size_t      field_detect_width(char const *buf, size_t blen);

/* Check (multibyte-safe) how many bytes of buf (which is blen byts) can be
 * safely placed in a buffer (field width) of maxlen bytes */
FL size_t      field_detect_clip(size_t maxlen, char const *buf, size_t blen);

/* Put maximally maxlen bytes of buf, a buffer of blen bytes, into store,
 * taking into account multibyte code point boundaries and possibly
 * encapsulating in bidi_info toggles as necessary */
FL size_t      field_put_bidi_clip(char *store, size_t maxlen, char const *buf,
                  size_t blen);

/* Place cp in a salloc()ed buffer, column-aligned; for header display only */
FL char *      colalign(char const *cp, int col, int fill,
                  int *cols_decr_used_or_null);

/* Convert a string to a displayable one;
 * prstr() returns the result savestr()d, prout() writes it */
FL void        makeprint(struct str const *in, struct str *out);
FL size_t      delctrl(char *cp, size_t len);
FL char *      prstr(char const *s);
FL int         prout(char const *s, size_t sz, FILE *fp);

/* Print out a Unicode character or a substitute for it, return 0 on error or
 * wcwidth() (or 1) on success */
FL size_t      putuc(int u, int c, FILE *fp);

/* Check wether bidirectional info maybe needed for blen bytes of bdat */
FL bool_t      bidi_info_needed(char const *bdat, size_t blen);

/* Create bidirectional text encapsulation information; without HAVE_NATCH_CHAR
 * the strings are always empty */
FL void        bidi_info_create(struct bidi_info *bip);

/*
 * wordexp.c
 */

/* Evaluate the string given as a new mailbox name. Supported meta characters:
 * . %  for my system mail box
 * . %user for user's system mail box
 * . #  for previous file
 * . &  invoker's mbox file
 * . +file file in folder directory
 * . any shell meta character (except for FEXP_NSHELL).
 * If FEXP_NSHELL is set you possibly want to call fexpand_nshell_quote(),
 * a poor man's vis(3), on name before calling this (and showing the user).
 * Returns the file name as an auto-reclaimed string */
FL char *      fexpand(char const *name, enum fexp_mode fexpm);

#define expand(N)                fexpand(N, FEXP_FULL)   /* XXX obsolete */
#define file_expand(N)           fexpand(N, FEXP_LOCAL)  /* XXX obsolete */

/* A poor man's vis(3) for only backslash escaping as for FEXP_NSHELL.
 * Returns the (possibly adjusted) buffer in auto-reclaimed storage */
FL char *      fexpand_nshell_quote(char const *name);

/* (Try to) Expand ^~/? and ^~USER/? constructs.
 * Returns the completely resolved (maybe empty or identical to input)
 * salloc()ed string */
FL char *      n_shell_expand_tilde(char const *s, bool_t *err_or_null);

/* (Try to) Expand any shell variable in s, allowing backslash escaping
 * (of any following character) with bsescape.
 * Returns the completely resolved (maybe empty) salloc()ed string.
 * Logs on error */
FL char *      n_shell_expand_var(char const *s, bool_t bsescape,
                  bool_t *err_or_null);

/* Check wether *s is an escape sequence, expand it as necessary.
 * Returns the expanded sequence or 0 if **s is NUL or PROMPT_STOP if it is \c.
 * *s is advanced to after the expanded sequence (as possible).
 * If use_prompt_extensions is set, an enum prompt_exp may be returned */
FL int         n_shell_expand_escape(char const **s,
                  bool_t use_prompt_extensions);

#ifndef HAVE_AMALGAMATION
# undef FL
# define FL
#endif

/* s-it-mode */
